#include "error.h"
#include "tree.h"
#include "asm.h"

#define FINITPUSH	4

static const char *instruction_name[Instruction::NUM_INSTRUCTIONS] = {
	"mov", "push", "pop", "add", "sub", "mul", "div", "call", "label",
	"jmp", "jmpz", "cmp", "sete", "setne", "setl", "setg", "setle",
	"setge", "and", "or", "xor", "ret"
};

void Assembly::ProcessInteger(const TreeNode &el) {
	ins.push_back(MOV(REG0, el.GetValue()));
}

void Assembly::ProcessString(const TreeNode &el) {
	ins.push_back(MOV(REG0, AT(RegisterString(el.GetName()))));
}

void Assembly::ProcessVariable(const TreeNode &el, bool address) {
	if (cur_function && cur_function->HasVar(el.GetName())) {
		ins.push_back(MOV(REG1, (cur_function->GetVarID(el.GetName()) + FINITPUSH) * sizeof(int)));
		ins.push_back(MOV(REG0, REGBP));
		ins.push_back(SUB(REG0, REG1));
		if (!address) {
			ins.push_back(MOV(REG0, AT(REG0)));
		}
	} else {
		if (address) {
			ins.push_back(MOV(REG0, AT(el.GetName())));
		} else {
			ins.push_back(MOV(REG0, el.GetName()));
		}
		CheckVar(el.GetName());
	}
}

void Assembly::PrepareOperands(const TreeNode &el) {
	ProcessCommand(el.GetSubNode(1));
	ins.push_back(PUSH(REG0));
	ProcessCommand(el.GetSubNode(0));
	ins.push_back(POP(REG1));
}

void Assembly::ProcessPoke(const TreeNode &el) {
	if (el.GetSubNodeCount() != 1)
		throw ERROR("Invalidni poke");
	ProcessCommand(el.GetSubNode(0));
}

void Assembly::ProcessEqual(const TreeNode &el) {
	if (el.GetSubNodeCount() != 2)
		throw ERROR("Invalidni equal");

	const TreeNode & lvalue = el.GetSubNode(0);

	switch (lvalue.GetType()) {
		case TreeNode::TN_var:
			if (cur_function && cur_function->HasVar(lvalue.GetName())) {
				ProcessVariable(lvalue, true);
				ins.push_back(PUSH(REG0));
				ProcessCommand(el.GetSubNode(1));
				ins.push_back(POP(REG1));
				ins.push_back(MOV(AT(REG1), REG0));
			} else {
				ProcessCommand(el.GetSubNode(1));
				ins.push_back(MOV(el.GetSubNode(0).GetName(), REG0));
				CheckVar(el.GetSubNode(0).GetName());
			}
			break;
		case TreeNode::TN_poke:
			ProcessPoke(lvalue);
			ins.push_back(PUSH(REG0));
			ProcessCommand(el.GetSubNode(1));
			ins.push_back(POP(REG1));
			ins.push_back(MOV(AT(REG1), REG0));
			break;
		default:
			throw ERROR("Invalidna l-vrijednost u stablu");
			break;
	}
}

void Assembly::ProcessOperator(const TreeNode &el) {
	const string &opr = el.GetName();

	if (opr.size() == 2 && opr[1] == '=') {
		char op = opr[0];

		PrepareOperands(el);
		ins.push_back(CMP(REG0, REG1));

		switch (op) {
			case '=':
				ins.push_back(SETE(REG0));
				break;
			case '>':
				ins.push_back(SETGE(REG0));
				break;
			case '<':
				ins.push_back(SETLE(REG0));
				break;
			case '!':
				ins.push_back(SETNE(REG0));
				break;
			default:
				throw ERROR("Los operator");
		}
		return;
	}

	if (opr.size() != 1)
		throw ERROR("Cudan operator \"" + opr + '"');

	switch (char op = opr[0]) {
		case '=':
			ProcessEqual(el);
			break;
		case '+':
			PrepareOperands(el);
			ins.push_back(ADD(REG0, REG1));
			break;
		case '-':
			PrepareOperands(el);
			ins.push_back(SUB(REG0, REG1));
			break;
		case '*':
			PrepareOperands(el);
			ins.push_back(MUL(REG0, REG1));
			break;
		case '/':
			PrepareOperands(el);
			ins.push_back(DIV(REG0, REG1));
			break;
		case '&':
			PrepareOperands(el);
			ins.push_back(AND(REG0, REG1));
			break;
		case '|':
			PrepareOperands(el);
			ins.push_back(OR(REG0, REG1));
			break;
		case '^':
			PrepareOperands(el);
			ins.push_back(XOR(REG0, REG1));
			break;
		case '>':
		case '<':
			PrepareOperands(el);
			ins.push_back(CMP(REG0, REG1));
			if (op == '>')	ins.push_back(SETG(REG0));
			else		ins.push_back(SETL(REG0));
			break;
		default:
			throw ERROR("Neocekivani operator: " + el.GetName());
	}
}

void Assembly::ProcessFunctionCall(const TreeNode &el) {
	if (el.GetSubNodeCount()) {
		vector<TreeNode>::const_iterator i = el.end();
		do {
			--i;
			ProcessCommand(*i);
			ins.push_back(PUSH(REG0));
		} while (i != el.begin());
	}

	if (el.GetValue() == TreeNode::CT_stdcall) {
		if (HasFunction(el.GetName())) {
			throw AsmError("Nije moguce pozvati svoju funkciju \"" + el.GetName() + "\" na stdcall nacin");
			return;
		}
		char argsz[40];
		sprintf(argsz, "@%lu", el.GetSubNodeCount() * sizeof(int));
		ins.push_back(CALL(el.GetName() + argsz));
	} else {
		ins.push_back(CALL(el.GetName()));
		if (el.GetSubNodeCount()) {
			ins.push_back(ADD(REGSP, el.GetSubNodeCount() * sizeof(int)));
		}
	}
}

void Assembly::ProcessFunction(const TreeNode &el) {
	AsmFunction &func = RegisterFunction(el.GetName(), !!el.GetValue());
	AsmFunction *prev_function = cur_function;
	cur_function = &func;

	ins.push_back(LABEL(el.GetName()));
	ins.push_back(MOV(REG0, REGSP));
	ins.push_back(PUSH(REG1));
	ins.push_back(PUSH(REG2));
	ins.push_back(PUSH(REG3));
	ins.push_back(PUSH(REGBP));
	ins.push_back(MOV(REGBP, REG0));

	ProcessBlock(el.begin(), el.end());

	ins.push_back(MOV(REG0, 0));

	ins.push_back(LABEL(func.GetEndLabel()));

	ins.push_back(MOV(REGSP, REGBP));
	ins.push_back(SUB(REGSP, FINITPUSH * sizeof(int)));
	ins.push_back(POP(REGBP));
	ins.push_back(POP(REG3));
	ins.push_back(POP(REG2));
	ins.push_back(POP(REG1));
	ins.push_back(RET());

	cur_function = prev_function;
}

void Assembly::ProcessWhile(const TreeNode &el) {
	const string &lbl1 = RegisterLabel("while_begin");
	const string &lbl2 = RegisterLabel("while_end");

	if (el.GetSubNodeCount() < 1)
		throw ERROR("Invalidan while nod");

	if (el.GetSubNodeCount() > 1) {
		ins.push_back(LABEL(lbl1));
		ProcessCommand(el.GetSubNode(0));
		ins.push_back(JMPZ(REG0, lbl2));
		ProcessBlock(el.begin() + 1, el.end());
		ins.push_back(JMP(lbl1));
		ins.push_back(LABEL(lbl2));
	} else {
		ProcessCommand(el.GetSubNode(0));
	}
}

void Assembly::ProcessIf(const TreeNode &el) {
	const string &lbl_else = RegisterLabel("else");
	const string &lbl_endif = RegisterLabel("endif");

	if (el.GetSubNodeCount() < 1)
		throw ERROR("Invalidan if nod");

	ProcessCommand(el.GetSubNode(0));
	ins.push_back(JMPZ(REG0, lbl_else));
	ProcessBlock(el.begin() + 1, el.begin() + el.GetValue() + 1);
	ins.push_back(JMP(lbl_endif));
	ins.push_back(LABEL(lbl_else));
	ProcessBlock(el.begin() + el.GetValue() + 1, el.end());
	ins.push_back(LABEL(lbl_endif));
}

void Assembly::ProcessArgument(const TreeNode &el) {
	if (!el.GetSubNodeCount())
		throw ERROR("Invalidan Argument nod");

	ProcessCommand(el.GetSubNode(0));
	ins.push_back(MUL(REG0, sizeof(int)));
	ins.push_back(ADD(REG0, REGBP));
	ins.push_back(MOV(REG0, AT(REG0)));
}

void Assembly::ProcessCommand(const TreeNode &el) {
	switch (el.GetType()) {
		case TreeNode::TN_operator:
			ProcessOperator(el);
			break;
		case TreeNode::TN_integer:
			ProcessInteger(el);
			break;
		case TreeNode::TN_string:
			ProcessString(el);
			break;
		case TreeNode::TN_var:
			ProcessVariable(el);
			break;
		case TreeNode::TN_varaddr:
			ProcessVariable(el, true);
			break;
		case TreeNode::TN_call:
			ProcessFunctionCall(el);
			break;
		case TreeNode::TN_function:
			ProcessFunction(el);
			break;
		case TreeNode::TN_return:
			if (!cur_function)
				throw ERROR("cur_function je NULL");
			ins.push_back(JMP(cur_function->GetEndLabel()));
			break;
		case TreeNode::TN_argument:
			ProcessArgument(el);
			break;
		case TreeNode::TN_while:
			ProcessWhile(el);
			break;
		case TreeNode::TN_if:
			ProcessIf(el);
			break;
		case TreeNode::TN_poke:
			ProcessPoke(el);
			ins.push_back(MOV(REG0, AT(REG0)));
			break;
		case TreeNode::TN_push:
			if (!cur_function)
				throw ERROR("Nema trenutne funkcije");
			cur_function->RegisterVariable(el.GetName());
			ins.push_back(PUSH(0));
			break;
		case TreeNode::TN_extern:
			extvars.push_back(el.GetName());
			break;
		default:
			throw ERROR("Neocekivani tip komande");
	}
}

void Assembly::ProcessBlock(teci a, teci b) {
	for (teci i = a; i != b; ++i) {
		ProcessCommand(*i);
	}
}

Assembly & Assembly::operator << (const Tree & t) {
	ProcessBlock(t.root.begin(), t.root.end());
	return *this;
}

void Assembly::CheckVar(const string &name) {
	if (!HasFunction(name) && !HasVariable(name) && !IsExtern(name)) {
                vars.push_back(name);
	}
}

string Assembly::RegisterString(const string &str) {
	char c_varlabel[100];
	sprintf(c_varlabel, "_str_%lu", strs.size() + 1);
	strs[c_varlabel] = str;
	return c_varlabel;
}

string Assembly::RegisterLabel(const string &templ) {
	string r;
	char *s = (char *)malloc(templ.size() + 40);
	sprintf(s, "lbl_%s_%d", templ.c_str(), ++lbls[templ]);
	r = s;
	free(s);
	return r;
}

AsmFunction & Assembly::RegisterFunction(const string &name, bool extrn) {
	if (HasFunction(name))
		throw AsmError("Funkcija s imenom \"" + name + "\" vec postoji");
	if (HasVariable(name))
		throw AsmError("Varijabla \"" + name + "\" smeta istoimenoj funkciji");

	AsmFunction &f = funcs[name] = AsmFunction(name);
	f.extrn = extrn;
	return f;
}

void Assembly::Dump() const {
	printf("Pseudo assembly dump:\n");
	for (vector<Instruction>::const_iterator i = ins.begin(); i != ins.end(); ++i) {
		i->Dump();
	}
}


string InstrOperand::ToStr() const {
	string r;
	if (!byval) r = '@';
	switch (reg) {
		case REG0: return r + "Reg0";
		case REG1: return r + "Reg1";
		case REG2: return r + "Reg2";
		case REG3: return r + "Reg3";
		case REGSP: return r + "StackPointer";
		case REGBP: return r + "BasePointer";
		case MEM: return r + name;
		case CONST: return "/const/";
		case UNDEFINED: return "-";
	}
	return "???";
}

string Instruction::GetName() const {
	return instruction_name[type];
}

void Instruction::Dump() const {
	printf("%s %s %s\n", instruction_name[type], a.ToStr().c_str(), b.ToStr().c_str());
}

AsmFunction::AsmFunction(const string &name) {
	_end_label = "__" + name + "_end";
	_name = name;
	extrn = false;
}

void AsmFunction::RegisterVariable(const string &name) {
	if (HasVar(name))
		throw AsmError("Lokalna varijabla " + name + " je vec definirana u " + GetName());
	vars[name] = vars.size();
}

int AsmFunction::GetVarID(const string &name) const {
	map<const string, int>::const_iterator i = vars.find(name);
	if (i == vars.end()) return -1;
	return i->second;
}
