#ifndef _ASM_H_
#define _ASM_H_

#include <string>
#include <vector>
#include <algorithm>
#include <map>
#include "tree.h"
#include "error.h"

using namespace std;

#define MOV(a, b) Instruction(Instruction::I_MOV, a, b)
#define PUSH(a) Instruction(Instruction::I_PUSH, a)
#define POP(a) Instruction(Instruction::I_POP, a)
#define ADD(a, b) Instruction(Instruction::I_ADD, a, b)
#define SUB(a, b) Instruction(Instruction::I_SUB, a, b)
#define MUL(a, b) Instruction(Instruction::I_MUL, a, b)
#define DIV(a, b) Instruction(Instruction::I_DIV, a, b)
#define CALL(a) Instruction(Instruction::I_CALL, a)
#define LABEL(a) Instruction(Instruction::I_LABEL, a)
#define JMPZ(a, b) Instruction(Instruction::I_JMPZ, a, b)
#define JMP(a) Instruction(Instruction::I_JMP, a)
#define CMP(a, b) Instruction(Instruction::I_CMP, a, b)
#define SETE(a) Instruction(Instruction::I_SETE, a)
#define SETNE(a) Instruction(Instruction::I_SETNE, a)
#define SETL(a) Instruction(Instruction::I_SETL, a)
#define SETG(a) Instruction(Instruction::I_SETG, a)
#define SETLE(a) Instruction(Instruction::I_SETLE, a)
#define SETGE(a) Instruction(Instruction::I_SETGE, a)
#define AND(a, b) Instruction(Instruction::I_AND, a, b)
#define OR(a, b) Instruction(Instruction::I_OR, a, b)
#define XOR(a, b) Instruction(Instruction::I_XOR, a, b)
#define RET() Instruction(Instruction::I_RET)

#define AT(x) InstrOperand::At(x)

typedef enum {
	REG0 = 0,
	REG1,
	REG2,
	REG3,
	REGSP,
	REGBP,
	REG_LAST = REGBP,
	MEM,
	CONST,
	UNDEFINED
} Register;

class InstrOperand {
	public:
		string name;
		Register reg;
		int cnst;
		bool byval;

		InstrOperand(const string & varname) {
			name = varname;
			reg = MEM;
			byval = true;
		}
		InstrOperand(Register _reg) {
			reg = _reg;
			byval = true;
		}
		InstrOperand(int c) {
			reg = CONST;
			cnst = c;
			byval = true;
		}

		string ToStr() const;
		static InstrOperand At(InstrOperand x) {
			x.byval = false;
			return x;
		}
};

class Instruction {
	public:
		typedef enum {
			I_MOV,
			I_PUSH,
			I_POP,
			I_ADD,
			I_SUB,
			I_MUL,
			I_DIV,
			I_CALL,
			I_LABEL,
			I_JMP,
			I_JMPZ,
			I_CMP,
			I_SETE,
			I_SETNE,
			I_SETL,
			I_SETG,
			I_SETLE,
			I_SETGE,
			I_AND,
			I_OR,
			I_XOR,
			I_RET,

			NUM_INSTRUCTIONS
		} Type;

		Type type;
		InstrOperand a, b;

		Instruction(Type t, InstrOperand _a = UNDEFINED, InstrOperand _b = UNDEFINED) :
			type(t), a(_a), b(_b) {}

		void Dump() const;
		string GetName() const;
};

class AsmFunction {
	public:
		bool extrn;

		AsmFunction() {}
		AsmFunction(const string &name);
		AsmFunction(const AsmFunction &c) {
			*this = c;
		}

		const string &GetName() const { return _name; }
		const string &GetEndLabel() const { return _end_label; }
		void RegisterVariable(const string &name);
		bool HasVar(const string &name) const { return vars.find(name) != vars.end(); }
		int GetVarID(const string &name) const;

	private:
		string _name;
		string _end_label;
		map<const string, int> vars;
};

class Assembly {
	public:
		void Dump() const;
		Assembly() : cur_function(NULL) {}

		Assembly & operator << (const Tree & t);
		const vector<Instruction> & GetInstructionList() const { return ins; }
		const vector<string> & GetVariableList() const { return vars; }
		const map<const string, string> & GetStringMap() const { return strs; }
		const map<const string, AsmFunction> & GetFunctionMap() const { return funcs; }
		bool IsExtern(const string &varname) const { return find(extvars.begin(), extvars.end(), varname) != extvars.end(); }

		bool HasFunction(const string &fname) const { return funcs.find(fname) != funcs.end(); }
		bool HasVariable(const string &vname) const { return find(vars.begin(), vars.end(), vname) != vars.end(); }

	private:
		typedef vector<TreeNode>::const_iterator teci;

		vector<Instruction> ins;
		vector<string> vars, extvars;
		map<const string, string> strs;
		map<const string, unsigned int> lbls;
		map<const string, AsmFunction> funcs;
		AsmFunction *cur_function;

		void ProcessVariable(const TreeNode &el, bool address=false);
		void ProcessOperator(const TreeNode &el);
		void ProcessCommand(const TreeNode &el);
		void ProcessBlock(teci a, teci b);
		void ProcessFunctionCall(const TreeNode &el);
		void ProcessFunction(const TreeNode &el);
		void ProcessArgument(const TreeNode &el);
		void ProcessInteger(const TreeNode &el);
		void ProcessString(const TreeNode &el);
		void ProcessWhile(const TreeNode &el);
		void ProcessIf(const TreeNode &el);
		void ProcessPoke(const TreeNode &el);
		void ProcessEqual(const TreeNode &el);

		void PrepareOperands(const TreeNode &el);
		void CheckVar(const string &name);
		string RegisterString(const string &str);
		string RegisterLabel(const string &templ);
		AsmFunction & RegisterFunction(const string &name, bool extrn);
};

class AsmError : public Error {
	public:
		AsmError(const string & desc) : Error("Assembling error: " + desc) {}
};

#endif
