#include "asm_linux.h"
#include "error.h"

#ifdef WIN32
#define F_PREFIX	"_"
#else
#define F_PREFIX	""
#endif

#define EXTENDBYTE()	fprintf(f, "\tmovzbl %s, %s\n", ByteOperand2S(i->a).c_str(), Operand2S(i->a).c_str())

static const char *bytereg[] = {"%al", "%bl", "%cl", "%dl"};
static const char *longreg[] = {"%eax", "%ebx", "%ecx", "%edx", "%esp", "%ebp"};

static string _Operand2S(const InstrOperand &op) {
	switch (op.reg) {
		case REG0:
		case REG1:
		case REG2:
		case REG3:
		case REGSP:
		case REGBP: return longreg[op.reg];
		case MEM: return F_PREFIX + op.name;
		case CONST:
			char buf[40];
			sprintf(buf, "$0x%X", op.cnst);
			return buf;
		case UNDEFINED:
		default:
			throw ERROR("Can't convert operand");
			return string();
	}
}

static string Operand2S(const InstrOperand &op) {
	string r;
	if (op.reg >= REG0 && op.reg <= REG_LAST) {
		r = _Operand2S(op);
		if (!op.byval) r = '(' + r + ')';
	} else {
		if (!op.byval) r += '$';
		r += _Operand2S(op);
	}
	return r;
}

static string ByteOperand2S(const InstrOperand &op) {
	string r;
	if (op.reg >= REG0 && op.reg <= REG3) {
		r = bytereg[op.reg];
		if (!op.byval) r='('+r+')';
	} else {
		throw ERROR("Nepodrzano");
	}
	return r;
}

bool GenerateLinuxAsm(const Assembly & a, FILE *f) {
	const vector<Instruction> &ins = a.GetInstructionList();
	const vector<string> &vars = a.GetVariableList();
	const map<const string, string> &strs = a.GetStringMap();
	const map<const string, AsmFunction> & funcs = a.GetFunctionMap();

	for (map<const string, AsmFunction>::const_iterator i = funcs.begin(); i != funcs.end(); ++i) {
		if (i->second.extrn)
			fprintf(f, ".globl " F_PREFIX "%s\n", i->first.c_str());
	}
	fputc('\n', f);

	fputs(".data\n", f);
	for (vector<string>::const_iterator i = vars.begin(); i != vars.end(); ++i) {
		if (!a.IsExtern(*i))
			fprintf(f, F_PREFIX "%s:\t.int 0\n", i->c_str());
	}
	fputc('\n', f);

	fputs(".text\n", f);

	fputs(F_PREFIX "dump:\n", f);
	fputs("\tpushl 4(%esp)\n", f);
	fputs("\tpushl $_dump_s\n", f);
	fputs("\tcall " F_PREFIX "printf\n", f);
	fputs("\taddl $8, %esp\n", f);
	fputs("\tmovl 4(%esp), %eax\n", f);
	fputs("\tret\n", f);
	fputs("_dump_s:\t.asciz \"%d\\n\"\n\n", f);


	for (vector<Instruction>::const_iterator i = ins.begin(); i != ins.end(); ++i) {
		switch (i->type) {
			case Instruction::I_MOV:
				fprintf(f, "\tmovl %s, %s\n", Operand2S(i->b).c_str(), Operand2S(i->a).c_str());
				break;
			case Instruction::I_PUSH:
				fprintf(f, "\tpushl %s\n", Operand2S(i->a).c_str());
				break;
			case Instruction::I_POP:
				fprintf(f, "\tpopl %s\n", Operand2S(i->a).c_str());
				break;
			case Instruction::I_ADD:
				fprintf(f, "\taddl %s, %s\n", Operand2S(i->b).c_str(), Operand2S(i->a).c_str());
				break;
			case Instruction::I_SUB:
				fprintf(f, "\tsubl %s, %s\n", Operand2S(i->b).c_str(), Operand2S(i->a).c_str());
				break;
			case Instruction::I_AND:
				fprintf(f, "\tandl %s, %s\n", Operand2S(i->b).c_str(), Operand2S(i->a).c_str());
				break;
			case Instruction::I_OR:
				fprintf(f, "\torl %s, %s\n", Operand2S(i->b).c_str(), Operand2S(i->a).c_str());
				break;
			case Instruction::I_XOR:
				fprintf(f, "\txorl %s, %s\n", Operand2S(i->b).c_str(), Operand2S(i->a).c_str());
				break;
			case Instruction::I_MUL:
				if (i->a.reg != REG0)
					fprintf(f, "\txchgl %%eax, %s\n", Operand2S(i->a).c_str());
				if (i->b.reg == CONST) {
					fprintf(f, "\tpushl %%ebx\n");
					fprintf(f, "\tmovl %s, %%ebx\n", Operand2S(i->b).c_str());
					fprintf(f, "\tmull %%ebx\n");
					fprintf(f, "\tpopl %%ebx\n");
				} else {
					fprintf(f, "\tmull %s\n", Operand2S(i->b).c_str());
				}
				if (i->a.reg != REG0)
					fprintf(f, "\txchgl %%eax, %s\n", Operand2S(i->a).c_str());
				break;
			case Instruction::I_DIV:
				fputs("\tpushl %edx\n", f);
				fputs("\txorl %edx, %edx\n", f);
				fprintf(f, "\tdivl %s, %s\n", Operand2S(i->b).c_str(), Operand2S(i->a).c_str());
				fputs("\tpopl %edx\n", f);
				break;
			case Instruction::I_CALL:
				fprintf(f, "\tcall %s\n", Operand2S(i->a).c_str());
				break;
			case Instruction::I_LABEL:
				fprintf(f, F_PREFIX "%s:\n", i->a.name.c_str());
				break;
			case Instruction::I_JMP:
				fprintf(f, "\tjmp %s\n", Operand2S(i->a).c_str());
				break;
			case Instruction::I_JMPZ:
				fprintf(f, "\torl %s, %s\n", Operand2S(i->a).c_str(), Operand2S(i->a).c_str());
				fprintf(f, "\tjz %s\n", Operand2S(i->b).c_str());
				break;
			case Instruction::I_CMP:
				fprintf(f, "\tcmp %s, %s\n", Operand2S(i->b).c_str(), Operand2S(i->a).c_str());
				break;
			case Instruction::I_SETE:
				fprintf(f, "\tsete %s\n", ByteOperand2S(i->a).c_str());
				EXTENDBYTE();
				break;
			case Instruction::I_SETNE:
				fprintf(f, "\tsetne %s\n", ByteOperand2S(i->a).c_str());
				EXTENDBYTE();
				break;
			case Instruction::I_SETL:
				fprintf(f, "\tsetl %s\n", ByteOperand2S(i->a).c_str());
				EXTENDBYTE();
				break;
			case Instruction::I_SETG:
				fprintf(f, "\tsetg %s\n", ByteOperand2S(i->a).c_str());
				EXTENDBYTE();
				break;
			case Instruction::I_SETLE:
				fprintf(f, "\tsetle %s\n", ByteOperand2S(i->a).c_str());
				EXTENDBYTE();
				break;
			case Instruction::I_SETGE:
				fprintf(f, "\tsetge %s\n", ByteOperand2S(i->a).c_str());
				EXTENDBYTE();
				break;
			case Instruction::I_RET:
				fprintf(f, "\tret\n");
				break;

			default:
				throw ERROR("Nedefinirana instrukcija \"" + i->GetName() + '"');
				return false;
		}
	}

	fputc('\n', f);

	for (map<const string, string>::const_iterator s = strs.begin(); s != strs.end(); ++s) {
		fprintf(f, F_PREFIX "%s:\t.asciz \"%s\"\n", s->first.c_str(), s->second.c_str());
	}

	return true;
}
