#include <stdio.h>
# ifdef WIN32
# include <windows.h>
# undef ERROR
# undef CONST
#else
# include <malloc.h>
# include <stdlib.h>
#endif

#include <vector>

#include "tokenizer.h"
#include "parser.h"
#include "error.h"
#include "asm.h"
#include "asm_linux.h"

using namespace std;

static void help(const char *exe) {
	puts("Chudan jezik kompajler");
	puts("    by DCihlar");
	printf("Uporaba: %s [args] infile\n", exe);
	puts("Argumenti:");
	puts(" -o<filename> = definira izlaznu datoteku");
	puts(" -S = izlaz je asemblerski kod");
	puts(" -v = verbose");
	puts(" -t = ispisi stablo");
	puts(" -a = ispisi pseudo asembler");
	puts(" -l = library za gcc");
	puts(" -L = library dir za gcc");
	puts(" -C<arg> = dodatni argument za gcc");
	puts(" -g<func> = ime funkcije za globalni kod (default: main)");
	puts(" -h = ovaj help");
}

static string GetCompiler() {
#ifdef WIN32
	char *t;
	if ((t = getenv("MINGWBIN")))
		return string(t) + "\\gcc.exe";
	if ((t = getenv("MINGWGCC")))
		return string(t);
#endif
	return "gcc";
}

class TempFile {
	public:
		TempFile(const string &templ, const char *mode = "wt") {
#ifdef WIN32
			char po[MAX_PATH];
			if (!GetTempFileName(".", templ.c_str(), 0, po))
				throw ERROR("GetTempFileName failed");
			fname = po;
			_filp = fopen(po, mode);
#else
			char *tempf = (char*)malloc(templ.size() + 7);
			sprintf(tempf, "%sXXXXXX", templ.c_str());
			_filp = fdopen(mkstemp(tempf), mode);
			fname = tempf;
			free(tempf);
#endif
			if (!_filp)
				throw ERROR("Ne mogu otvoriti temp datoteku");
		}

		~TempFile() {
			Close();
			if (!fname.empty()) unlink(fname.c_str());
		}

		void Close() {
			if (_filp) {
				fclose(_filp);
				_filp = NULL;
			}
		}

		FILE *GetFilp() { return _filp; }
		const string & GetName() const { return fname; }

	private:
		FILE *_filp;
		string fname;
};

int main(int argn, const char *argv[]) {
	Tokenizer tok;
	Assembly a;
	Parser par;
	FILE *f = NULL;
	string src_file, out_file;
	enum {
		OF_EXE,
		OF_ASM
	} oformat = OF_EXE;
	bool do_tree = false, do_pasm = false, verbose = false;
	vector<string> gcc_addargs;

	for (int i = 1; i < argn; ++i) {
		if (*argv[i] == '-' && argv[i][1]) {
			switch (argv[i][1]) {
				case 'S':
					oformat = OF_ASM;
					break;
				case 'o':
					out_file = argv[i]+2;
					break;
				case 'v':
					verbose = true;
					par.SetVerbose(true);
					break;
				case 't':
					do_tree = true;
					break;
				case 'a':
					do_pasm = true;
					break;
				case 'l':
					gcc_addargs.push_back(argv[i]);
					break;
				case 'L':
					gcc_addargs.push_back(argv[i]);
					break;
				case 'C':
					gcc_addargs.push_back(string("-") + (argv[i]+2));
					break;
				case 'g':
					par.SetRootFunctionName(argv[i]+2);
					break;

				case 'h':
					help(argv[0]);
					return 0;

				default:
					help(argv[0]);
					return 1;
			}
			continue;
		}

		if (src_file.empty()) src_file = argv[i];
		else {
			fprintf(stderr, "Vec je ulazna datoteka!\n");
			return 1;
		}
	}

	if (src_file.empty()) {
		fprintf(stderr, "Nije definirana ulazna datoteka!\n");
		return 1;
	}

	if (out_file.empty()) {
		out_file = "a";
		switch (oformat) {
			case OF_EXE:
#ifdef WIN32
				out_file += ".exe";
#else
				out_file += ".out";
#endif
				break;
			case OF_ASM:
				out_file += ".s";
				break;
		}
	}

	try {

		if (src_file == "-") {
			f = stdin;
		} else {
			f = fopen(src_file.c_str(), "rt");
			if (!f) {
				fprintf(stderr, "Datoteka s izvornim kodom nije nadjena!\n");
				return 1;
			}
		}
		tok.Read(f);
		if (f != stdin) fclose(f);
		f = NULL;

		par << tok;

		a << par.tree;

		if (do_tree) {
			par.tree.Dump();
		}

		if (do_pasm) {
			a.Dump();
		}

		switch (oformat) {
			case OF_EXE:
				{
				TempFile f("asm_tmp");
				GenerateLinuxAsm(a, f.GetFilp());
				f.Close();

				string buildcmd = GetCompiler();
				buildcmd += " -x assembler ";
				buildcmd += string(f.GetName());
				buildcmd += " -o" + out_file;
				for (vector<string>::const_iterator i = gcc_addargs.begin(); i != gcc_addargs.end(); ++i)
					buildcmd += ' ' + *i;
				if (verbose) {
					printf("Compiling with:\n%s\n", buildcmd.c_str());
				}
				if (system(buildcmd.c_str())) {
					fprintf(stderr, "gcc je fejlao!\n");
					return 1;
				}
				break;
				}
			case OF_ASM:
				f = fopen(out_file.c_str(), "wt");
				if (!f) {
					fprintf(stderr, "Ne mogu otvoriti izlaznu datoteku!\n");
					return 1;
				}
				GenerateLinuxAsm(a, f);
				fclose(f);
				break;
		}

	} catch (const ParseError & e) {
		e.Dump();
		par.GetCurrentToken().Dump();
		if (f) fclose(f);
		return 1;
	} catch (const Error & e) {
		e.Dump();
		if (f) fclose(f);
		return 1;
	}

	return 0;
}
