#ifndef _TOKENIZER_H_
#define _TOKENIZER_H_

#include <stdio.h>
#include <string>
#include <vector>

#include "error.h"

using namespace std;


class Token {
	public:
		typedef enum {
			T_unknown,
			T_keyword,
			T_number,
			T_string,
			T_var,
			T_operator,
			T_block,
			T_semicolon,
			T_bracket,
			T_comma,
			T_address,
			T_end
		} Type;

		Token(Type type = T_unknown) : _type(type), _integer(-1), _line(-1) {}

		bool Read(FILE *f, int line);
		void Dump() const;
		int GetLine() const { return _line; }
		string GetData() const { return _data; }
		int GetInt() const { return _integer; }

		bool IsOperator(const char *op) const { return _type == T_operator && op == _data; }
		bool IsOperator(char op) const { return _type == T_operator && _data.size() == 1 && _data[0] == op; }
		bool IsOperator() const { return _type == T_operator; }
		bool IsConst() const { return _type == T_number || _type == T_string; }
		bool IsNumConst() const { return _type == T_number; }
		bool IsStrConst() const { return _type == T_string; }
		bool IsVar() const { return _type == T_var; }
		bool IsBracket(char c) const { return _type == T_bracket && _data[0] == c; }
		bool IsBracket() const { return _type == T_bracket; }
		bool IsSemicolon() const { return _type == T_semicolon; }
		bool IsEnd() const { return _type == T_end; }
		bool IsKeyword() const { return _type == T_keyword; }
		bool IsKeyword(const string &name) const { return _type == T_keyword && name == _data; }
		bool IsComma() const { return _type == T_comma; }
		bool IsAddress() const { return _type == T_address; }
		bool IsBlock(char c) const { return _type == T_block && _data[0] == c; }

	private:
		Type _type;
		string _data;
		int _integer;
		int _line;

		bool SkipSpaces(FILE *f);
		static string ReadString(FILE *f, char type);
};

class CTokenStream : public vector<Token>::const_iterator {
	public:
		bool verbose;

		CTokenStream() : verbose(false) {};
		CTokenStream(const vector<Token>::const_iterator &p) : vector<Token>::const_iterator(p), verbose(false) {}

		CTokenStream & operator ++ () {
			++*((vector<Token>::const_iterator*)this);
			if (verbose)
				(*this)->Dump();
			return *this;
		}

		CTokenStream & operator ++ (int) {
			++*((vector<Token>::const_iterator*)this);
			if (verbose)
				(*this)->Dump();
			return *this;
		}
};

class Tokenizer {
	public:
		void Read(FILE *f);
		CTokenStream GetStream() const;

		void Dump() const;

	private:
		vector<Token> tokens;
};

class TokenError : public Error {
	public:
		TokenError(const string &desc) : Error("Token error: " + desc) {}
};


extern const char *keywords[];
extern const char *functions[];

#endif
